@@ -1,4 +1,5 @@
 import argparse
+import importlib.metadata
 from textwrap import dedent
 from unittest.mock import patch
 
@@ -7,7 +8,6 @@ import pytest
 
 import mdformat
 from mdformat._cli import run
-from mdformat._compat import importlib_metadata
 from mdformat.plugins import (
     _PARSER_EXTENSION_DISTS,
     CODEFORMATTERS,
@@ -172,39 +172,6 @@ def test_table(monkeypatch):
     )
 
 
-class ExamplePluginWithCli:
-    """A plugin that adds CLI options."""
-
-    @staticmethod
-    def update_mdit(mdit: MarkdownIt):
-        mdit.enable("table")
-
-    @staticmethod
-    def add_cli_options(parser: argparse.ArgumentParser) -> None:
-        parser.add_argument("--o1", type=str)
-        parser.add_argument("--o2", type=str, default="a")
-        parser.add_argument("--o3", dest="arg_name", type=int)
-
-
-def test_cli_options(monkeypatch, tmp_path):
-    """Test that CLI arguments added by plugins are correctly added to the
-    options dict."""
-    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExamplePluginWithCli)
-    file_path = tmp_path / "test_markdown.md"
-    file_path.touch()
-
-    with patch.object(MDRenderer, "render", return_value="") as mock_render:
-        assert run((str(file_path), "--o1", "other", "--o3", "4")) == 0
-
-    (call_,) = mock_render.call_args_list
-    posargs = call_[0]
-    # Options is the second positional arg of MDRender.render
-    opts = posargs[1]
-    assert opts["mdformat"]["o1"] == "other"
-    assert opts["mdformat"]["o2"] == "a"
-    assert opts["mdformat"]["arg_name"] == 4
-
-
 class ExamplePluginWithGroupedCli:
     """A plugin that adds CLI options."""
 
@@ -272,6 +239,35 @@ dont_override_toml = 'dont override this with None if CLI opt not given'
     )
 
 
+def test_plugin_argument_warnings(monkeypatch, tmp_path):
+    """Test for warnings of plugin arguments that conflict with TOML."""
+
+    class ExamplePluginWithStoreTrue:
+        @staticmethod
+        def update_mdit(mdit: MarkdownIt):
+            pass
+
+        @staticmethod
+        def add_cli_argument_group(group: argparse._ArgumentGroup) -> None:
+            group.add_argument("--store-true", action="store_true")
+            group.add_argument("--store-false", action="store_false")
+            group.add_argument("--store-zero", default=0)
+            group.add_argument("--store-const", action="store_const", const=True)
+
+    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExamplePluginWithStoreTrue)
+    file_path = tmp_path / "test_markdown.md"
+    file_path.touch()
+
+    with patch.object(MDRenderer, "render", return_value=""):
+        with pytest.warns(DeprecationWarning) as warnings:
+            assert run([str(file_path)]) == 0
+
+    assert "--store-true" in str(warnings.pop().message)
+    assert "--store-false" in str(warnings.pop().message)
+    assert "--store-zero" in str(warnings.pop().message)
+    assert len(warnings) == 0
+
+
 def test_cli_options_group__no_toml(monkeypatch, tmp_path):
     """Test add_cli_argument_group plugin API with configuration only from
     CLI."""
@@ -410,7 +406,7 @@ Name: mdformat-gfm
 Version: 0.3.6
 """
     )
-    distro = importlib_metadata.PathDistribution(dist_info_path)
+    distro = importlib.metadata.PathDistribution(dist_info_path)
     entrypoints = distro.entry_points
 
     loaded_eps, dist_infos = _load_entrypoints(entrypoints)
